@@ -1,19 +1,17 @@
 from __future__ import annotations
 
 import asyncio
+import logging
 import re
 from typing import Any, Callable, Dict, List, Optional, Tuple
 
 import click
 
 from gptsh.core.config_api import compute_tools_policy
+from gptsh.core.exceptions import ReplExit
 from gptsh.mcp import ensure_sessions_started_async as ensure_sessions_started_async  # noqa: F401
 
 
-class ReplExit(Exception):
-    pass
-
-
 def build_prompt(
     *,
     agent_name: Optional[str],
@@ -79,6 +77,9 @@ def command_reasoning_effort(
     return agent_conf
 
 
+_log = logging.getLogger(__name__)
+
+
 def command_tools(agent: Any) -> str:
     """Return a formatted list of tools for the current agent.
 
@@ -97,8 +98,8 @@ def command_tools(agent: Any) -> str:
             try:
                 if policy and policy.is_auto_allowed(server, name):
                     badge = " \u2714"  # checkmark for auto-approved
-            except Exception:
-                pass
+            except Exception as e:
+                _log.debug("policy.is_auto_allowed failed for %s/%s: %s", server, name, e)
             lines.append(f"  - {name}{badge}")
     return "\n".join(lines)
 
@@ -204,7 +205,8 @@ def command_agent(
     try:
         nonce = (mcp_cfg.get("_repl_nonce") or 0) + 1
         mcp_cfg["_repl_nonce"] = nonce
-    except Exception:
+    except Exception as e:
+        _log.debug("Failed to bump MCP nonce: %s", e)
         mcp_cfg["_repl_nonce"] = 1
     mgr = None
     prompt_str = build_prompt(
@@ -251,27 +253,54 @@ def command_help() -> str:
 
 
 def setup_readline(get_agent_names: Callable[[], List[str]]) -> Tuple[bool, Any]:
-    """Configure GNU readline with a simple completer for REPL slash-commands.
+    """Configure readline/libedit with a simple completer for REPL slash-commands.
     Returns (enabled, readline_module_or_None).
+
+    Notes:
+    - On macOS Python is often linked against libedit instead of GNU readline.
+      In that case the correct binding for tab completion is
+      "bind ^I rl_complete" instead of "tab: complete".
     """
     try:
-        import readline as _readline  # type: ignore
-    except Exception:
+        try:
+            import gnureadline as _readline  # type: ignore
+        except ImportError:
+            import readline as _readline  # type: ignore
+    except Exception as e:
+        _log.warning("readline import failed: %s", e)
         return False, None
     try:
-        _readline.parse_and_bind("tab: complete")
+        try:
+            doc = getattr(_readline, "__doc__", "") or ""
+            if "libedit" in doc.lower():
+                _log.debug("readline: libedit detected")
+                # macOS/libedit: different binding syntax for tab completion
+                _readline.parse_and_bind("bind ^I rl_complete")
+                # Enable incremental reverse search on Ctrl-R (and forward on Ctrl-S)
+                _readline.parse_and_bind("bind ^R em-inc-search-prev")
+                _readline.parse_and_bind("bind ^S em-inc-search-next")
+            else:
+                # GNU readline
+                _readline.parse_and_bind("tab: complete")
+                # Enable reverse/forward history search
+                _readline.parse_and_bind('"\\C-r": reverse-search-history')
+                _readline.parse_and_bind('"\\C-s": forward-search-history')
+        except Exception as e:
+            # Best-effort: log and continue to attempt to set completer
+            _log.debug("readline parse_and_bind failed: %s", e)
         try:
             delims = _readline.get_completer_delims()
             if "/" in delims:
                 _readline.set_completer_delims(delims.replace("/", ""))
-        except Exception:
-            pass
+        except Exception as e:
+            _log.debug("failed to adjust completer delimiters: %s", e)
         commands = get_command_names()
 
         def _completer(text, state):
             try:
                 buf = _readline.get_line_buffer()
-            except Exception:
+            except Exception as e:
+                _log.debug("readline.get_line_buffer failed: %s", e)
                 buf = ""
             if not buf.startswith("/"):
                 return None
@@ -284,13 +313,15 @@ def setup_readline(get_agent_names: Callable[[], List[str]]) -> Tuple[bool, Any]
             arg_prefix = ""
             try:
                 arg_prefix = "" if buf.endswith(" ") else (text or "")
-            except Exception:
+            except Exception as e:
+                _log.debug("computing arg_prefix failed: %s", e)
                 arg_prefix = text or ""
             if cmd == "/agent":
                 names = []
                 try:
                     names = list(get_agent_names() or [])
-                except Exception:
+                except Exception as e:
+                    _log.debug("get_agent_names failed: %s", e)
                     names = []
                 opts = [n for n in names if n.startswith(arg_prefix)]
                 return opts[state] if state < len(opts) else None
@@ -305,10 +336,11 @@ def setup_readline(get_agent_names: Callable[[], List[str]]) -> Tuple[bool, Any]
 
         try:
             _readline.set_completer(_completer)
-        except Exception:
-            pass
+        except Exception as e:
+            _log.debug("set_completer failed: %s", e)
         return True, _readline
-    except Exception:
+    except Exception as e:
+        _log.warning("setup_readline failed: %s", e)
         return False, None
 
 
@@ -317,8 +349,8 @@ def add_history(readline_module: Any, line: str) -> None:
         return
     try:
         readline_module.add_history(line)
-    except Exception:
-        pass
+    except Exception as e:
+        _log.debug("add_history failed: %s", e)
 
 
 
@@ -366,13 +398,20 @@ async def run_agent_repl_async(
     # Heuristic: if agent has no tools, disable tools in non-stream flow
     try:
         no_tools = not any(len(v or []) > 0 for v in (agent.tools or {}).values())
-    except Exception:
+    except Exception as e:
+        _log.debug("Failed to inspect agent tools: %s", e)
         no_tools = True
 
     # Keep a heuristic no_tools flag based on resolved agent tools
 
     history_messages: List[Dict[str, Any]] = []
     last_interrupt = 0.0
+    # Persistent MCP manager for the REPL session (reused across turns)
+    try:
+        from gptsh.mcp.manager import MCPManager as _MCPManager
+    except Exception:  # pragma: no cover - fallback
+        _MCPManager = None  # type: ignore
+    mcp_manager = (None if no_tools or _MCPManager is None else _MCPManager(config))
 
     async def _run_once(user_text: str) -> str:
         # Reuse CLI run_llm to ensure consistent streaming fallback and tool behavior
@@ -392,37 +431,41 @@ async def run_agent_repl_async(
             history_messages=history_messages,
             result_sink=sink,
             agent_obj=agent,
+            mcp_manager=mcp_manager,
         )
         return (sink[0] if sink else "")
 
-    if initial_prompt and str(initial_prompt).strip():
-        try:
-            user_msg = {"role": "user", "content": initial_prompt}
-            content = await _run_once(initial_prompt)
-            history_messages.extend([user_msg, {"role": "assistant", "content": content}])
-        except KeyboardInterrupt:
-            last_interrupt = time.monotonic()
-            click.echo("Cancelled.", err=True)
-
     while True:
-        try:
-            line = input(prompt_str)
-        except KeyboardInterrupt:
-            now = time.monotonic()
-            if now - last_interrupt <= 1.5:
+        if initial_prompt:
+            line = initial_prompt
+            initial_prompt = None
+        else:
+            try:
+                doc = getattr(rl, "__doc__", "") or ""
+                if "libedit" in doc.lower():
+                    # MacOS non-GNU readline compatibility
+                    click.echo(prompt_str, nl=False)
+                    line = input()
+                else:
+                    line = input(prompt_str)
+            except KeyboardInterrupt:
+                now = time.monotonic()
+                if now - last_interrupt <= 1.5:
+                    click.echo("", err=True)
+                    break
+                last_interrupt = now
+                click.echo("(^C) Press Ctrl-C again to exit", err=True)
+                continue
+            except EOFError:
                 click.echo("", err=True)
                 break
-            last_interrupt = now
-            click.echo("(^C) Press Ctrl-C again to exit", err=True)
-            continue
-        except EOFError:
-            click.echo("", err=True)
-            break
 
         sline = line.strip()
         if not sline:
             continue
 
+        add_history(rl, sline)
+
         if sline.startswith("/"):
             parts = sline.split(None, 1)
             cmd = parts[0]
@@ -449,8 +492,8 @@ async def run_agent_repl_async(
                 cli_model_override = new_override
                 try:
                     agent.llm._base["model"] = cli_model_override
-                except Exception:
-                    pass
+                except Exception as e:
+                    _log.debug("Failed to update agent base model: %s", e)
                 provider_conf_local["model"] = cli_model_override
                 prompt_str = new_prompt
                 continue
@@ -492,13 +535,18 @@ async def run_agent_repl_async(
                     cli_model_override = model
                     provider_conf_local["model"] = model
                     prompt_str = prompt_out
+                    # Update MCP manager based on new no_tools state
+                    if _MCPManager is not None:
+                        mcp_manager = (None if no_tools else (_MCPManager(config)))
                 except Exception as e:
+                    _log.warning("Failed to switch agent: %s", e)
                     click.echo(f"Failed to switch agent: {e}", err=True)
                 continue
             if cmd == "/tools":
                 try:
                     click.echo(command_tools(agent))
                 except Exception as e:
+                    _log.warning("Failed to list tools: %s", e)
                     click.echo(f"Failed to list tools: {e}", err=True)
                 continue
             if cmd == "/no-tools":
@@ -518,17 +566,22 @@ async def run_agent_repl_async(
                     agent = new_agent
                     no_tools = _no
                     click.echo(msg)
+                    # Update persistent MCP manager
+                    if _MCPManager is not None:
+                        mcp_manager = (None if no_tools else (_MCPManager(config)))
                 except Exception as e:
+                    _log.warning("Failed to toggle tools: %s", e)
                     click.echo(f"Failed to toggle tools: {e}", err=True)
                 continue
             click.echo("Unknown command", err=True)
             continue
 
-        add_history(rl, sline)
         try:
             user_msg = {"role": "user", "content": sline}
             content = await _run_once(sline)
-            history_messages.extend([user_msg, {"role": "assistant", "content": content}])
+            # ChatSession/run_llm already appends assistant/tool messages into history_messages.
+            # We only need to record the user's message for this turn.
+            history_messages.append(user_msg)
         except KeyboardInterrupt:
             last_interrupt = time.monotonic()
             click.echo("Cancelled.", err=True)
